/**
 ******************************************************************************
 * @file           : main.c
 * @author         : Auto-generated by STM32CubeIDE
 * @brief          : Main program body
 ******************************************************************************
 * @attention
 *
 * Copyright (c) 2022 STMicroelectronics.
 * All rights reserved.
 *
 * This software is licensed under terms that can be found in the LICENSE file
 * in the root directory of this software component.
 * If no LICENSE file comes with this software, it is provided AS-IS.
 *
 ******************************************************************************
 */

#include <stdint.h>
#include <stdio.h>

/* Debug Exception and Monitor Control Register base address */
#define DEMCR                 *((volatile uint32_t*) 0xE000EDFCu)
/* ITM register addresses */
#define ITM_STIMULUS_PORT0    *((volatile uint32_t*) 0xE0000000u)
#define ITM_TRACE_EN          *((volatile uint32_t*) 0xE0000E00u)
/* System Handler control and state register */
#define SHCSR                 *((volatile uint32_t*) 0xE000ED24u)
/* SysTick register address */
#define SCSR                  *((volatile uint32_t*) 0xE000E010u)
#define SRVR                  *((volatile uint32_t*) 0xE000E014u)
/* Interrupt Control and State Register */
#define ICSR                  *((volatile uint32_t*) 0xE000ED04u)

/* Send a char through ITM */
void ITM_SendChar(uint8_t ch) {
    // read FIFO status in bit [0]:
    while(!(ITM_STIMULUS_PORT0 & 1));
    // write to ITM stimulus port0
    ITM_STIMULUS_PORT0 = ch;
}

/* Override low-level _write system call */
int _write(int file, char *ptr, int len) {
    int DataIdx;
    for (DataIdx = 0; DataIdx < len; DataIdx++) {
        ITM_SendChar(*ptr++);
    }
    return len;
}

/* RAM */
#define RAM_START         (0x20000000u)
#define RAM_SIZE          (128 * 1024) // 128 KB

/* Stacks */
#define MAIN_STACK        (RAM_START + RAM_SIZE)
#define TASK_NUMBER_MAX   (16)
#define TASK_STACK_SIZE   (1024u)

typedef enum {
  RUNNING,
  BLOCKED
} TaskState_t;

typedef struct {
  void (*handler)(void);
  TaskState_t state;
  uint32_t psp;
  uint32_t waitToTick;
} TCB_t;

uint32_t __gSysTicks = 0;
uint32_t __uCurrentTaskIdx = 0;
TCB_t __pTasks[TASK_NUMBER_MAX] = {0};

/* Scheduler */
uint32_t get_current_psp() {
  return __pTasks[__uCurrentTaskIdx].psp;
}

void save_current_psp(uint32_t psp) {
  __pTasks[__uCurrentTaskIdx].psp = psp;
}

void select_next_task() {
  /* Check all task state */
  for(int i=0; i<TASK_NUMBER_MAX; i++) {
    if(__pTasks[i].state == BLOCKED) {
      if(__gSysTicks >= __pTasks[i].waitToTick) {
        __pTasks[i].state = RUNNING;
      }
    }
  }

  /* Round-Robin scheduler */
  while(1) {
    __uCurrentTaskIdx++;
    // check if a task is register at current slot
    if (__uCurrentTaskIdx >= TASK_NUMBER_MAX || __pTasks[__uCurrentTaskIdx].psp == 0) {
      __uCurrentTaskIdx=0;
    }

    if(__pTasks[__uCurrentTaskIdx].state == RUNNING) break;
  }
}

void idle_main(void) {
  while(1) {
    __asm volatile("NOP");
  }
}

void init_task(void (*handler)) {

  // init idle if it is not found
  if (handler != idle_main) {
    if (__pTasks[0].handler != idle_main) {
      init_task(idle_main);
    }
  }

  // find an empty slot
  int i=0;
  for(; i<TASK_NUMBER_MAX; i++) {
    if (__pTasks[i].psp == 0) break;
  }


  if(i >= TASK_NUMBER_MAX) {
    printf("Can not register a new task anymore!\n");
    return;
  } else {
    printf("Register a task %p at slot %i\n", handler, i);
  }


  // calculate new PSP
  uint32_t* psp = (uint32_t*)(MAIN_STACK - (i+1)*TASK_STACK_SIZE);

  // fill dummy stack frame
  *(--psp) = 0x01000000u; // Dummy xPSR, just enable Thumb State bit;
  *(--psp) = (uint32_t) handler; // PC
  *(--psp) = 0xFFFFFFFDu; // LR with EXC_RETURN to return to Thread using PSP
  *(--psp) = 0x12121212u; // Dummy R12
  *(--psp) = 0x03030303u; // Dummy R3
  *(--psp) = 0x02020202u; // Dummy R2
  *(--psp) = 0x01010101u; // Dummy R1
  *(--psp) = 0x00000000u; // Dummy R0
  *(--psp) = 0x11111111u; // Dummy R11
  *(--psp) = 0x10101010u; // Dummy R10
  *(--psp) = 0x09090909u; // Dummy R9
  *(--psp) = 0x08080808u; // Dummy R8
  *(--psp) = 0x07070707u; // Dummy R7
  *(--psp) = 0x06060606u; // Dummy R6
  *(--psp) = 0x05050505u; // Dummy R5
  *(--psp) = 0x04040404u; // Dummy R4

  // save PSP
  __pTasks[i].psp = (uint32_t)psp;

  // initial state
  __pTasks[i].state = RUNNING;
  __pTasks[i].handler = handler;
}

void start_scheduler() {
  printf("Start Scheduler!\n");

  // start with the first task
  __uCurrentTaskIdx = 0;

  // prepare PSP of the first task
  __asm volatile("BL get_current_psp"); // return PSP in R0
  __asm volatile("MSR PSP, R0");  // set PSP

  // change to use PSP
  __asm volatile("MRS R0, CONTROL");
  __asm volatile("ORR R0, R0, #2"); // set bit[1] SPSEL
  __asm volatile("MSR CONTROL, R0");

  // start SysTick
  // clear and set the period
  SRVR &= ~0xFFFFFFFF;
  SRVR |= 16000-1; // 1000 Hz ~ 1 ms
  // enable SysTick
  SCSR |= (1 << 1); // enable SysTick Exception request
  SCSR |= (1 << 2); // select system clock
  SCSR |= (1 << 0); // start

  // Move to Unprivileged level
  __asm volatile("MRS R0, CONTROL");
  __asm volatile("ORR R0, R0, #1"); // Set bit[0] nPRIV
  __asm volatile("MSR CONTROL, R0");
  // right after here, access is limited

  // get the handler of the first task by tracing back from PSP which is at R4 slot
  void (*handler)() = __pTasks[__uCurrentTaskIdx].handler;

  // execute the handler
  handler();
}

void reschedule(uint32_t priv) {
  if(priv) {
    // trigger PendSV directly
    ICSR |= (1 << 28);
  } else {
    // call Supervisor exception to get Privileged access
    __asm volatile("SVC #255");
  }
}

void task_delay(uint32_t ticks) {
  if(__uCurrentTaskIdx) {
    __pTasks[__uCurrentTaskIdx].state = BLOCKED;
    __pTasks[__uCurrentTaskIdx].waitToTick = __gSysTicks + ticks;
    reschedule(0); // unprivileged
  }
}

/* Context Switching run here */
void SysTick_Handler(void) {
  printf("****\n");
  __gSysTicks++;
  reschedule(1); // privileged
}

/* Switch Context */
__attribute__ ((naked)) void PendSV_Handler(void) {
  // save LR back to main, must do this firstly
   __asm volatile("PUSH {LR}");

  /* Save the context of current task */

  // get current PSP
  __asm volatile("MRS R0, PSP");
  // save R4 to R11 to PSP Frame Stack
  __asm volatile("STMDB R0!, {R4-R11}"); // R0 is updated after decrement
  // save current value of PSP
  __asm volatile("BL save_current_psp"); // R0 is first argument

  /* Do scheduling */

  // select next task
  __asm volatile("BL select_next_task");

  /* Retrieve the context of next task */

  // get its past PSP value
  __asm volatile("BL get_current_psp"); // return PSP is in R0
  // retrieve R4-R11 from PSP Fram Stack
  __asm volatile("LDMIA R0!, {R4-R11}"); // R0 is updated after increment
  // update PSP
  __asm volatile("MSR PSP, R0");

  // exit
  __asm volatile("POP {LR}");
  __asm volatile("BX LR");
}

__attribute__ ((naked)) void SVC_Handler(void) {
  __asm volatile("TST LR, 4"); // check LR to know which stack is used
  __asm volatile("ITE EQ"); // 2 next instructions are conditional
  __asm volatile("MRSEQ R0, MSP"); // save MSP if bit 2 is 0
  __asm volatile("MRSNE R0, PSP"); // save PSP if bit 2 is 1
  __asm volatile("B SVC_Handler_main"); // pass R0 as the argument
}

void SVC_Handler_main(uint32_t* SP) {
  // get the address of the instruction saved in PC
  uint8_t *pInstruction = (uint8_t*)(SP[6]);

  // go back 2 bytes (16-bit opcode)
  pInstruction -= 2;

  // get the opcode, in little endian
  uint8_t svc_num = *pInstruction;

  switch(svc_num) {
    case 0xFF:
      // trigger PendSV
      ICSR |= (1 << 28);
      break;
    default:
      break;
  }
}

/* Fault Handlers */
void HardFault_Handler() {
  printf("Exception: HardFault_Handler\n");
  while(1);
}

void MemManage_Handler() {
  printf("Exception: MemManage_Handler\n");
  while(1);
}

void BusFault_Handler() {
  printf("Exception: BusFault_Handler\n");
  while(1);
}

void UsageFault_Handler() {
  printf("Exception: Usage Fault\n");
  while(1);
}

/* Tasks */
void task1_main(void) {
  while(1) {
    printf("1111\n");
    task_delay(16);
  }
}

void task2_main(void) {
  while(1) {
    printf("2222\n");
    task_delay(8);
  }
}

void task3_main(void) {
  while(1) {
    printf("3333\n");
    task_delay(4);
  }
}

void task4_main(void) {
  while(1) {
    printf("4444\n");
    task_delay(2);
  }
}

int main(void)
{
  // Enable TRCENA
  DEMCR |= ( 1 << 24);
  // Enable stimulus port 0
  ITM_TRACE_EN |= ( 1 << 0);

  /***************
   * Fault traps *
   ***************/
  // Enable Fault Handlers
  SHCSR |= (1 << 16); // Memory Fault
  SHCSR |= (1 << 17); // Bus Fault
  SHCSR |= (1 << 18); // Usage Fault

  /*************
   * SCHEDULER *
   *************/
  init_task(task1_main);
  init_task(task2_main);
  init_task(task3_main);
  init_task(task4_main);

  start_scheduler();

  // should never go here
  for(;;);
}
